/*
 * Animator Library
 * Copyright (C) 2021-2024  Karl Robillard
 * SPDX-License-Identifier: GPL-3.0-or-later
 */

#ifndef ANIM_H
#define ANIM_H

#include <stdint.h>

enum AnimState {
    ANIM_FREE,
    ANIM_PLAYING,
    ANIM_PAUSED
};

enum CurveType {
    CURVE_LINEAR,
    CURVE_EASE_IN,
    CURVE_EASE_OUT,
    CURVE_EASE_IO
};

#define ANIM_FOREVER    0xffff

typedef uint16_t AnimId;
#define ANIM_UNUSED     0xffff

typedef struct {
    void* bank;
    //void (*finish)(void*, uint32_t);
    //void* finishData;
    uint32_t avail;
    uint32_t used;
    uint32_t firstFree;
}
Animator;

inline float clampSumF(float a, float b, float low, float high)
{
    a += b;
    return (a < low) ? low : (a > high ? high : a);
}

#ifdef __cplusplus
extern "C" {
#endif

int    anim_init(Animator*, int count, void (*finishFunc)(void*, uint32_t),
                 void* fdata);
void   anim_free(Animator*);
void   anim_clear(Animator*);
void   anim_clearGroup(Animator* an, uint16_t group, int flush);
void   anim_advance(Animator*, float seconds);
AnimId anim_startCycleRandomI(Animator* an, uint16_t group, float dur,
                              int loops, int start, int end, int chance,
                              int32_t* dest);
AnimId anim_startI1(Animator* an, uint16_t group, int curve, float dur,
                    int32_t start, int32_t end, int32_t* dest);
AnimId anim_startF1(Animator* an, uint16_t group, int curve, float dur,
                    float start, float end, float* dest);
AnimId anim_startF2(Animator* an, uint16_t group, int curve, float dur,
                    const float* start, const float* end, float* dest);
AnimId anim_startF3(Animator* an, uint16_t group, int curve, float dur,
                    const float* start, const float* end, float* dest);
void   anim_setState(Animator*, AnimId, int animState);
int32_t anim_valueI(const Animator*, AnimId);
float* anim_valueF2(const Animator*, AnimId);

#ifdef __cplusplus
}
#endif

#endif // ANIM_H
