/*
 * Gypsy Road - Configuration
 * Copyright (C) 2023-2024  Karl Robillard
 * SPDX-License-Identifier: GPL-3.0-or-later
 */

#ifndef CONFIG_H
#define CONFIG_H

#include "gypsy_types.h"
#include "module.h"
#include "stage.h"

#define CONF_MODULE 1

enum ResourceType {
    RT_MUSIC,
    RT_SOUND,
    RT_FONT,
    RT_IMAGE,
    RT_TEXTURE,
    RT_FRAMEBUF,
    RT_VARRAY,
    RT_SHADER,
    RT_TILEMAP,
    RT_COUNT
};

typedef struct {
    uint16_t resType;
    uint16_t rbank;
    uint16_t bankIndex;
    uint16_t cdiIndex;
    uint16_t specIt;
    uint16_t specEnd;
} ResourceInfo;


#define BANK_LIMIT  5

typedef struct {
    Module mod;
    uint16_t resBanks[RT_COUNT + RT_COUNT * BANK_LIMIT];
} Config;  

struct CreatureDef;
struct Image32;
struct ImageInfo;
struct LocaleDef;

Config* conf_init(const char* module, const char* soundtrack);
void    conf_free(Config*);
int     conf_bankSize(int resType);
int     conf_loadInitialResources();
int     conf_loadResources(const Symbol* names, int count,
                           uint16_t* slotArr, int* error);
const ResourceInfo* conf_resource(Symbol name);
const ResourceInfo* conf_resourceQuery(Symbol name);
int     conf_enterStage(int stage);
void    conf_internSymbols(Symbol*, uint16_t limit, const char* names);
const char*     conf_symbolName(Symbol);
const char*     conf_string(StringId);
const char*     conf_modulePath(const CDIEntry*);
void*           conf_loadChunk(const CDIEntry*);
const CDIEntry* conf_fileEntry(const char* filename);
const CDIEntry* conf_musicFile(uint32_t id);
const CDIEntry* conf_soundFile(uint32_t id);
const float*    conf_voiceParts(uint32_t id);
int             conf_loadImageChunk(uint32_t cdiIndex, struct Image32*);
int             conf_loadImage(Symbol name, struct ImageInfo*);
#ifdef USE_XU4
const void*     conf_egaPalette();
int     conf_localeType(int id);
bool    conf_localeDef(struct LocaleDef*, int id);
const struct CreatureDef* conf_creatureDef(int ctype);
const struct TerrainDef*  conf_terrainDef(int tileId);
#endif

#endif  // CONFIG_H
