#include <stdint.h>
#include "image32.h"

enum GpuDrawList {
    GPU_DLIST_GUI,
    GPU_DLIST_HUD,
    GPU_DLIST_VIEW_OBJ,
    GPU_DLIST_VIEW_FX
};

enum GpuClutValues {
    COL_BLACK    = 0,
    COL_WHITE    = 1,
    COL_TX_BLACK = 2,
    COL_SD_BROWN = 4,
    COL_DK_BROWN = 5,
    COL_ORANGE   = 7,
    COL_PL_ORANGE = 8,
    COL_BEIGE    = 9,
    COL_PL_YELLOW = 10,
    COL_BROWN    = 17,
    COL_YELLOW   = 19,
    COL_MINT     = 32,
    COL_LT_GREEN = 33,
    COL_CYAN     = 42,
    COL_LT_BLUE  = 44,
    COL_BLUE     = 46,
    COL_PURPLE   = 54,
    COL_LT_GRAY  = 63,
    COL_DK_GRAY  = 66,
    COL_TRANS    = 128
};

#define WID_NONE    -1

#define GPU_TEX_MINF_LINEAR 1
#define GPU_TEX_MIPF_LINEAR 2
#define GPU_TEX_MINF_MIP    4
#define GPU_TEX_MAGF_LINEAR 8
#define GPU_TEX_DEPTH       16

#define GPU_TEX_NEAREST     0
#define GPU_TEX_LINEAR      9

#define GPU_ATTR_COUNT      7

typedef struct WorkRegion WorkRegion;
typedef struct WorkBuffer WorkBuffer;

struct WorkRegion {
    uint32_t start;
    uint32_t avail;
    uint32_t used;
};

struct WorkBuffer {
    float*   attr;
    uint32_t dirty;         // LIMIT: 32 regions.
    uint16_t regionCount;
    WorkRegion region[2];
};

typedef struct {
    uint32_t colorTex;
    uint16_t w;
    uint16_t h;
} GpuFramebufInfo;


WorkBuffer* gpu_allocWorkBuffer(const int* regionSizes, int regionCount);
void        gpu_freeWorkBuffer(WorkBuffer*);
float*      gpu_beginRegion(WorkBuffer*, int regionN);
float*      gpu_appendRegion(WorkBuffer*, int regionN);
void        gpu_endRegion(WorkBuffer* work, int regionN, float* attr);

const char* gpu_init(void* res, int w, int h, int scale, int filter);
void     gpu_free(void* res);
void     gpu_adjustDisplay(void* res, int w, int h, int scale, int filter);
void     gpu_viewport(int x, int y, int w, int h);
const GpuFramebufInfo* gpu_framebufInfo(int slot);
const GpuFramebufInfo* gpu_useFramebuf(int slot);
uint32_t gpu_shaderProgram(int slot);
uint32_t gpu_vertexArray(int slot);
uint32_t gpu_vertexBuffer(int slot);
uint32_t gpu_texture(int slot);
uint32_t gpu_makeTexture(const Image32* img);
void     gpu_blitTexture(uint32_t tex, int x, int y, const Image32* img);
void     gpu_freeTexture(uint32_t id);
uint32_t gpu_screenTexture(void* res);
void     gpu_drawTextureScaled(void* res, uint32_t tex, int invert);
void     gpu_drawViewportQuad(void* res);
void     gpu_clear(void* res, const float* color);
void     gpu_invertColors(void* res);
void     gpu_setScissor(int* box);
void     gpu_updateWorkBuffer(void* res, int list, WorkBuffer*);
void     gpu_updateWorkBufferA(int vaSlot, WorkBuffer*);
void     gpu_drawQuadsWork(void* res, int list, const WorkBuffer*, int mask);
void     gpu_drawQuadsRegion(void* res, int list, const WorkRegion*);
void     gpu_drawQuadsRegionA(int vaSlot, int fpv, const WorkRegion*);
float*   gpu_beginAttr(void* res, int list);
void     gpu_endAttr(void* res, int list, float* attr);
void     gpu_clearAttr(void* res, int list);
void     gpu_drawTris(void* res, int list);
void     gpu_drawQuads(void* res, int list);
void     gpu_enableGui(void* res, int wid, int mode, int iconsTexSlot);
void     gpu_guiClutUV(void* res, float* uv, float colorIndex);
void     gpu_guiSetOrigin(void* res, float x, float y);
float*   gpu_emitQuad(float* attr, const float* drawRect, const float* uvRect);
float*   gpu_emitQuadPq(float* attr, const float* drawRect, const float* uvRect,
                        float posZ, float texP, float texQ);
float*   gpu_emitSkinQuads(float* attr, const float* rect, const float* uvs,
                           float cornerW, float cornerH, const float* texSize);
float*   gpu_emitHSkinQuads(float* attr, const float* rect, const float* uvs,
                            float cornerW, const float* texSize);

struct TileMap;
struct TileMapView;
struct BlockingGroups;

void     gpu_resetTileMap(void* res, const struct TileMap* map);
void     gpu_useTileMapShader(void* res, const struct TileMapView* view,
                              const struct BlockingGroups* blocks, int invert);
void     gpu_drawTileMap(void* res, const struct TileMapView* view);
int      gpu_renderTinyMap(void* res, struct TileMapView* view, Image32* img);
