#ifdef __APPLE__
#include <OpenGL/gl.h>
#elif defined(__ANDROID__)
#include <GLES3/gl31.h>
#elif defined(_WIN32)
#include "glad.h"
#else
#define GL_GLEXT_PROTOTYPES
#include <GL/gl.h>
#endif

#include "tilemap.h"

enum VertexArrayDefOpcode {
    VA_END,
    VA_ELEM16,      // Static indices.
    VA_QUAD_ELEM16, // Static indices for quads.
    VA_SINGLE,
    VA_DOUBLE,
    VA_STATIC,      // Single-buffered, static
    VA_LAYOUT       // Interleaved attribute sizes (VA_ATTR)
};

#define VA_ATTR2(a,b)       ((b<<3) | a)
#define VA_ATTR3(a,b,c)     ((c<<6) | (b<<3) | a)
#define VA_ATTR4(a,b,c,d)   ((d<<9) | (c<<6) | (b<<3) | a)
#define VA_EBUF             0x8000

enum GLObject {
    GLOB_GUI_LIST,      // GPU_DLIST_GUI (single-buffered)
    GLOB_QUAD,          // No DrawList
    GLOB_HUD_LIST0,     // GPU_DLIST_HUD
    GLOB_HUD_LIST1,
    GLOB_DRAW_LIST0,    // GPU_DLIST_VIEW_OBJ
    GLOB_DRAW_LIST1,
    GLOB_FX_LIST0,      // GPU_DLIST_VIEW_FX
    GLOB_FX_LIST1,
#ifdef USE_TILEMAP
    GLOB_MAPFX_LIST0,
    GLOB_MAPFX_LIST1,
    GLOB_MAP_CHUNK0,
    GLOB_MAP_CHUNK1,
    GLOB_MAP_CHUNK2,
    GLOB_MAP_CHUNK3,
#endif
    GLOB_COUNT
};

enum GLTextureUnit {
    GTU_CMAP,
    GTU_MATERIAL,
    GTU_NOISE,
    GTU_SHADOW,
    GTU_SCALER_LUT
};

typedef struct {
    uint16_t bufI;      // GLObject vbo/vao index toggle.
    uint8_t  dual;      // Double-buffered.
    uint8_t  fpv;       // Floats per vertex.
    int     byteSize;
    GLsizei count;      // Number of floats.
} DrawList;

#define CHUNK_FX_LIMIT  8

struct MapFx {
    float x, y, w, h;
    float u, v, u2, v2;
    //AnimId anim;
};
typedef struct MapFx MapFx;

#define BUFFER_COUNT   GLOB_COUNT+1
#define QUAD_ELEM_BUF  GLOB_COUNT
#define TEXTURE_COUNT  4

typedef struct {
    void* resBank;

    GLuint screenTex;
    GLuint whiteTex;
    GLuint noiseTex;
    GLuint shadowTex;
#ifdef SHADOWCAST
    GLuint shadowFbo;
#endif
    GLuint fontTex;
    GLuint guiTex;
    float guiTexSize[2];
    GLuint tilesTex;
    GLuint materialTex;

    GLuint vbo[ BUFFER_COUNT ];
    GLuint vao[ GLOB_COUNT ];

    int    activeFilter;
    GLuint scalerLut;
    GLuint scaler;
    GLint  slocScMat;
    GLint  slocScDim;
    GLint  slocScOut;

    GLuint shadeColor;
    GLint  cmapTrans;

    GLuint shadeSolid;
    GLint  solidTrans;
    GLint  solidColor;

    GLuint shadeGlyph;
    GLint  glyphTrans;
    GLint  glyphOrigin;
    GLint  glyphWidget;
    GLint  glyphFg;

#ifdef SHADOWCAST
    GLuint shadow;
    GLint  shadowTrans;
    GLint  shadowVport;
    GLint  shadowViewer;
    GLint  shadowCounts;
    GLint  shadowShapes;
#endif

    GLuint shadeWorld;
    GLint  worldTrans;
    GLint  worldOrigin;
    GLint  worldShadowMap;
    GLint  worldScroll;
    GLint  worldCatalog;

    DrawList dl[5];
    float* dptr;
#ifdef USE_TILEMAP
    const TileMap* tmap;
    //const TileRenderData* renderData;
    int    blockCount;
    GLsizei mapChunkVertCount;
    GLsizei mapChunkElemCount[4];
    uint16_t mapChunkId[4];     // Chunk X,Y of associated GLOB_MAP_CHUNK.
    uint16_t mapChunkFxUsed[4];
    MapFx mapChunkFx[4*CHUNK_FX_LIMIT];
#endif
} OpenGLResources;

extern const char* ogl_genResources(GLuint* names, const uint8_t* rcount,
                             const uint16_t* verteArrayDef,
                             const char** shaderFiles);
extern void ogl_deleteResources(GLuint* names, const uint8_t* rcount);
