#ifndef IMAGE_H
#define IMAGE_H

#include "config.h"
#include "image32.h"

enum U4Filetype {
    FTYPE_UNKNOWN,
    FTYPE_U4RAW,
    FTYPE_U4RLE,
    FTYPE_U4LZW,
    FTYPE_U5LZW,
    FTYPE_FMTOWNS,
    FTYPE_FMTOWNS_PIC,
    FTYPE_FMTOWNS_TIF
};

typedef struct {
    Symbol name;
    int16_t x, y, width, height;
    int16_t hotX, hotY;
    uint16_t celCount;
} ImageRect;

struct ImageInfo {
    Image32 img;
    ImageRect* rect;
    int16_t rectCount;
    uint16_t tileDim;
};
typedef struct ImageInfo ImageInfo;

enum AtlasEditOpcode {
    AEDIT_NOP,
    AEDIT_BRUSH,
    AEDIT_RECT,
    AEDIT_OP_COUNT
};

typedef struct {
    uint16_t name;      // Image name or AtlasEditOpcode.
    int16_t x, y, w, h;
} AtlasSubImage;

//#define IMAGE_BOTTOM_0
#define BPP_CLUT8   -8

#define image_init(inf)         memset(inf, 0, sizeof(ImageInfo))
#define image_freeRectQ(inf)    free((inf)->rect)

Image32* loadImage_png(FILE *file, Image32* image);
Image32* loadImage_webp(FILE *file, Image32* image);
Image32* loadImage_u4(const char* filename, int ftype, int width, int height,
                      int bpp, Image32* image);
int      saveImage_png(FILE *file, Image32* image, int invert);
void image_free(ImageInfo*);
void image_freeRect(ImageInfo*);
void image_buildAtlas(ImageInfo* atlas, int atlasW, int atlasH,
                      const AtlasSubImage* asiBuffer, int count);
int  image_findRect(const ImageInfo*, Symbol name);
const ImageInfo* image_lookupRect(Symbol name, int* rectN);
void image_rectUV(const ImageInfo*, int rectN, float* uvs);

#endif
