#ifndef PRESENT_H
#define PRESENT_H

#include <boron/urlan.h>
#include "tilemap.h"
#include "txf_draw.h"

typedef struct {
    const uint8_t* tileTable;
    const uint8_t* code;
    uint32_t pc;
    uint32_t wait;
} U4SpriteProgram;

typedef struct {
    uint16_t flags;
    uint16_t func;
    uint16_t frameCount;
    uint16_t frameDur;      // Milliseconds.
    uint16_t frameN;
    uint16_t texInfoN;
    uint16_t rectN;
    uint16_t _pad0;
    float x, y, w, h;       // Kept together as gpu_emit* rectangle.
    float s0, t0, s1, t1;   // Kept together as gpu_emit* uvs.
    float z, p, q;
    float funcVar;
    uint32_t frameFlipTime;
    const uint8_t* frameTable;
} PresentSprite;

typedef struct {
    PresentSprite* data;
    size_t used, avail;
} PSpriteArray;

#define RENDER_PROG_SLOTS 4

enum PresentControlFlags {
    PRESENT_CTRL_OVERSKIP = 1,
    PRESENT_CTRL_NOSOUND  = 2,
    PRESENT_CTRL_NOMUSIC  = 4,
    PRESENT_CTRL_ANYKEY   = 8
};

typedef struct {
    const uint32_t* icode;
    const uint32_t* pc;
    UCell* varyCells;
    const UCell* resourceTable;
    const UCell* renderWords;
    const UCell* renderTable;
    uint8_t   renderProgSlot[RENDER_PROG_SLOTS];
    uint32_t  pipe;
    uint32_t  waitTimeout;
    UIndex    linkedBlkN;
    uint16_t  control;
    int16_t   stageId;
    int       fbTarget;         // Framebuffer resource bank slot.
    uint32_t  randJitterMask;
    uint32_t  randJitter;
    uint32_t  spriteSheet;      // GL texture name
    float     screenCam[16];
    float     targetCam[16];
    PSpriteArray sprites;
    TxfDrawState textState;
    uint16_t  textVaSlot;
    int16_t   textCursorSprite;
    uint32_t  textVtxCount;
    float     textPos[2];
#ifdef USE_TILEMAP
    TileMapView tmapView;
#endif
    U4SpriteProgram u4Prog;     // TODO: Make this generic.
} PresentProgram;

#define present_disable(prog)   (prog)->pc = NULL

void present_free(PresentProgram*);
void present_link(UThread*, PresentProgram*, const UCell* fbTarget,
                  UIndex specBlkN);
void present_reset(PresentProgram*);
void present_screenResized(PresentProgram*);
void present_stopVary(PresentProgram*);
int  present_ready(PresentProgram*);
int  present_run(PresentProgram*);
void present_render(PresentProgram*);
void present_control(PresentProgram*, int flags);

#endif  // PRESENT_H
