#ifndef SCREEN_H
#define SCREEN_H

#include "config.h"
#include "settings.h"

struct TxfHeader;

typedef enum {
    SCR_FORMAT_FIXED,
    SCR_FORMAT_RESIZABLE,
    SCR_FORMAT_SCALED
} ScreenFormatG;

typedef struct {
    const struct TxfHeader* const* fontTable;
    ScreenFormatG format;
    int displayW;       // Full display pixel dimensions.
    int displayH;
    int aspectW;        // Aspect-correct pixel dimensions.
    int aspectH;
    int aspectX;        // Origin of aspect-correct area on display.
    int aspectY;
    uint16_t initW, initH;
    uint16_t slocFullscreen;
    uint16_t slocWindowSize;
    uint16_t slocScale;
    uint16_t slocFilter;
    uint16_t slocUseJoystick;
} ScreenBase;

typedef enum {
    MC_DEFAULT,
    MC_WEST,
    MC_NORTH,
    MC_EAST,
    MC_SOUTH
} MouseCursor;

ScreenBase* screen_init(ScreenFormatG fmt, int rw, int rh, const SettingsBase*);
void    screen_free(ScreenBase*);
void    screen_swap(ScreenBase*);
int     screen_loadFont(Symbol name, const CDIEntry* ent, int slot);
void    screen_setMouseCursor(MouseCursor cursor);
void    screen_showMouseCursor(bool visible);
void    screen_handleEvents(ScreenBase*);
void    screen_toggleFullscreen();

#endif  // SCREEN_H
