#ifndef SETTINGS_H
#define SETTINGS_H

#include <stdbool.h>
#include <stdint.h>

typedef struct SettingsBase    SettingsBase;
typedef struct SettingsDef     SettingsDef;

typedef struct {
    uint16_t count;
    uint16_t loc;
} SettingsGroup;

struct SettingsBase {
    const char* appName;
    const char* appDirectory;
    const char* appVersion;
    SettingsGroup* sections;
    uint16_t sectionCount;
    uint16_t dirty;
};

enum SettingsDataType {
    SDT_UNSET,
    SDT_BOOL,
    SDT_ENUM,
    SDT_INT,
    SDT_FLOAT,
    SDT_COORD,
    SDT_STRING
};

struct SettingsDef {
    const char* name;
    uint8_t  dataType;
    uint8_t  hidden;
    uint16_t loc;
    union {
        struct { uint16_t count, secOffset; } en;
        struct { int32_t min, max, step; } i;
        struct { float min, max, step; } f;
        struct { int16_t min[2], max[2]; } c;
    } range;
};

#define INVALID_SETTING 0x8000

SettingsBase* settings_init(const char* appName, const char* appDir,
                            const char* appVersion, const uint8_t* spec,
                            size_t specLen);
void     settings_free(SettingsBase*);
int      settings_path(const SettingsBase*, StringTable* path);
void     settings_sections(SettingsBase*, StringTable* stab);
void     settings_location(uint16_t* sloc, uint16_t limit, const char* names);
bool     settings_bool(uint16_t location);
int      settings_int(uint16_t location);
float    settings_float(uint16_t location);
const int16_t* settings_coord(uint16_t location);
const char* settings_string(uint16_t location);
bool     settings_pokeB(uint16_t location, bool val);
bool     settings_pokeI(uint16_t location, int val);
bool     settings_pokeF(uint16_t location, float val);
bool     settings_pokeC(uint16_t location, int16_t x, int16_t y);
bool     settings_pokeS(uint16_t location, const char* val);
uint16_t settings_next(uint16_t location);
bool     settings_def(uint16_t location, SettingsDef* def);
void     settings_reset();

#endif
