#ifndef TILEMAP_H
#define TILEMAP_H

#define USE_TILEMAP
#ifdef USE_TILEMAP

#include "gypsy_types.h"

typedef uint16_t TileId;

struct TileMap {
    Symbol name;
    uint16_t w, h, d;
    uint16_t chunkW, chunkH;
    uint16_t boundMaxX, boundMaxY;
    TileId* grid;
};

enum TileFlags {
    TILE_FLIPH  = 0x1000,
    TILE_FLIPV  = 0x2000,
    TILE_HIDDEN = 0x4000,
    TILE_HOLE   = 0x8000,
    TILE_INDEX  = 0x0fff
};

#define BLOCKING_POS_SIZE   128*3

struct BlockingGroups {
    int left, center, right;
    float tilePos[BLOCKING_POS_SIZE];
};

struct TileVisualEffect {
    float x, y;
    int32_t frame;
    uint16_t method;
    uint16_t animId;    // AnimId
    TileId tile;
};

enum TileVisualEffectMethod {
    TVE_FREE,
    TVE_SPRITE,
    TVE_SPRITE_MOVE,
    TVE_HIDE = 0x8000
};

typedef struct TileMap TileMap;
typedef struct TileMapForms TileMapForms;
typedef struct TileVisualEffect TileVisualEffect;
typedef struct BlockingGroups BlockingGroups;

struct TileMapForms {
    TileVisualEffect* data;
    uint32_t used, avail;
    int firstFree;
};

struct TileMapView {
    int viewport[4];        // Screen rectangle (x, y, width, height)
    float cx, cy;           // Map position at center of viewport.
    float scale;            // 1.0 = 1 texel/pixel
    float tileUnitW, tileUnitH;
    float catalogCols;      // Number of tiles per row.
    float catalogRows;      // Number of tile rows.
    uint16_t catalogSlot;   // Texture resource for tile catalog (diffuse).
    //uint16_t materialSlot;  // Texture resource for tile material.
    uint16_t tilemapSlot;   // Tilemap resource.
    uint16_t tileDim;       // Size in pixels.
    uint16_t catalogColsI;  // Number of tiles per row (integer).
    uint16_t animGroup;
    TileId   hiddenTile;    // Replaces grid entries where TILE_HIDDEN is set.

    uint16_t effectAvail;
    uint16_t effectUsed;
    TileVisualEffect* effect;   // A bank of user managed effects.

    TileMapForms forms;         // Forms for objects on map.
};

typedef struct TileMapView TileMapView;

const uint8_t* tilemap_loadU4(TileMap*, const uint8_t* data,
                              uint16_t borderTile);
void     tilemap_free(int slot);
//const TileMap* tilemap_lookup(Symbol name);
bool     tilemap_outOfBounds(const TileMap*, int x, int y);

void tview_initEffects(TileMapView* view, int effectCount, int formCount,
                       uint16_t animGroup);
void tview_freeEffects(TileMapView*);
void tview_setCatalog(TileMapView*, int textureSlot);
void tview_viewScreen(TileMapView*);
void tview_tileAt(const TileMapView*, int mouseX, int mouseY, int* tileCoord);
int  tview_tileUVs(const TileMapView*, TileId tile, float* uv);
void tview_clearEffects(TileMapView*);
TileVisualEffect* tview_useEffect(TileMapView*, int effectN, TileId tile,
                                  float x, float y);
void tview_removeEffect(TileMapView*, int effectN);
TileVisualEffect* tview_addForm(TileMapView*); //TileId tile, float x, float y);
void tview_removeForm(TileMapView*, TileVisualEffect*);
void tview_animateCreature(const TileMapView*, TileVisualEffect*,
                           uint16_t ctype, int initFrame);

#endif
#endif
