/*
 * Gypsy Example - Introduction
 */

#include "gypsy.h"
#include "hid_key.h"
#include "present.h"
#include "stages.h"

// STAGE_INTRO
void intro_enter(Stage* st, uintptr_t args)
{
    (void) args;

    present_disable(&gs.present);
    if (! conf_enterStage(stage_id(st)))
        stage_exit();
}

void intro_leave(Stage* st)
{
    (void) st;
    present_stopVary(&gs.present);
}


int game_globalKey(Stage* st, const InputEvent* ie);

int intro_dispatch(Stage* st, const InputEvent* ie)
{
    if (ie->type == IE_FRAME_UPDATE) {
        PresentProgram* prog = &gs.present;
        if (present_ready(prog))
            present_run(prog);
        present_render(prog);
    } else if (ie->type == IE_CYCLE_TIMER) {
        anim_advance(&gs.varySt, (float) gs.timerInterval * 0.001f);
    } else if (ie->type == IE_KEY_PRESS) {
        switch (ie->key.uid) {
            case HKEY_J:
                stage_run(STAGE_PLAY);
                return 1;
#ifdef DEBUG
            case HKEY_Backspace:
                // Restart presentation.
                present_stopVary(&gs.present);
                present_reset(&gs.present);
                return 1;
#endif
        }
        if (! game_globalKey(st, ie) && ie->key.uid <= HKEY_KP_Delete) {
            if (gs.present.control & PRESENT_CTRL_ANYKEY)
                gs.present.control &= ~PRESENT_CTRL_ANYKEY;
            else
                present_control(&gs.present, PRESENT_CTRL_OVERSKIP |
                                             PRESENT_CTRL_NOSOUND);
        }
        return 1;
    }
    return 0;
}
