/*
 * Gypsy Example - Play Game
 */

#include <stdlib.h>
#include "gypsy.h"
#include "hid_key.h"
#include "present.h"
#include "stages.h"

// Game world state.
typedef struct {
    int cycleCount;
} Simulation;

// STAGE_PLAY
void play_enter(Stage* st, uintptr_t args)
{
    (void) args;

    present_disable(&gs.present);
    if (! conf_enterStage(stage_id(st)))
        stage_exit();
}

void play_leave(Stage* st)
{
    (void) st;
    present_stopVary(&gs.present);
}

int game_globalKey(Stage* st, const InputEvent* ie)
{
    (void) st;

    switch (ie->key.uid) {
    case HKEY_F:
        if (ie->key.modifier & IE_MOD_CTRL) {
            screen_toggleFullscreen();
            return 1;
        }
        break;
    case HKEY_Q:
        if (ie->key.modifier & IE_MOD_CTRL) {
            stage_exit();
            return 1;
        }
        break;
    }
    return 0;
}

int play_dispatch(Stage* st, const InputEvent* ie)
{
    if (ie->type == IE_FRAME_UPDATE) {
        PresentProgram* prog = &gs.present;
        if (present_ready(prog))
            present_run(prog);
        present_render(prog);
    } else if (ie->type == IE_CYCLE_TIMER) {
        ((Simulation*) gs.gameSim)->cycleCount++;
        anim_advance(&gs.varySt, (float) gs.timerInterval * 0.001f);
    } else if (ie->type == IE_KEY_PRESS) {
        switch (ie->key.uid) {
            case HKEY_I:
                stage_run(STAGE_INTRO);
                return 1;
        }
        if (! game_globalKey(st, ie) && ie->key.uid <= HKEY_KP_Delete) {
            if (gs.present.control & PRESENT_CTRL_ANYKEY)
                gs.present.control &= ~PRESENT_CTRL_ANYKEY;
            else
                present_control(&gs.present, PRESENT_CTRL_OVERSKIP |
                                             PRESENT_CTRL_NOSOUND);
        }
        return 1;
    }
    return 0;
}

/*
 * NotifyHandler for SENDER_GAME messages.
 */
void game_notice(int sender, void* message, void* user)
{
    (void) user;

    if (sender == SENDER_GAME) {
        Simulation* sim;
        switch (GAME_MESSAGE_TYPE) {
            case GAME_INIT:
                gs.gameSim = sim = (Simulation*) calloc(1, sizeof(Simulation));
                sim->cycleCount = 0;
                break;

            case GAME_FREE:
                sim = (Simulation*) gs.gameSim;
                printf("Simulation cycles: %d\n", sim->cycleCount);
                free(gs.gameSim);
                gs.gameSim = NULL;
                break;

            case GAME_RESET:
            case GAME_SAVE:
            case GAME_LOAD:
                break;
        }
    }
}
