#!/opt/gypsy_sdk-0.1.2/bin/boron -s
; Create Gypsy module package.
; Requires Boron 2.1.1.

usage: {{
Usage: pack-gypsy [<OPTIONS>] <module-path>

Options:
  -h           Print this help and exit.
  -o <module>  Output module file.
  -v <level>   Set verbose level. (0-2, default is 1)
  --version    Print version and exit.
}}

root-path: %module/Ultima-IV/
module-file: none
verbose: 1

forall args [
	switch first args [
		"-v" [verbose: to-int second ++ args]
		"-h" [print usage quit]
		"-o" [module-file: to-file second ++ args]
		"--version" [print "pack-gypsy 0.1.0" quit]
		[root-path: terminate to-file first args '/']
	]
]

fatal: func ['code msg] [
	print msg
	quit/return select [
		usage	64  ; EX_USAGE
		data	65  ; EX_DATAERR
		noinput	66  ; EX_NOINPUT
		config	78  ; EX_CONFIG
	] code
]

path-slash: charset "/\"
basename: func [path] [
	if pos: find/last path path-slash [return next pos]
	path
]

ifn module-file [
	module-file: join last split root-path path-slash %.mod
]
if gt? size? basename module-file 39 [
	fatal config "Module filename must be less than 40 characters"
]

enum-value: func [list val] [
	either pos: find list val [index?/zero pos] 0
]

file-ext: func [file] [find/last file '.']

class-constraint: context [
	mage:   0x01  bard:    0x02  fighter: 0x04  druid:    0x08
	tinker: 0x10  paladin: 0x20  ranger:  0x40  shepherd: 0x80
]

; Spec is ['not block! | 'only block!]
make-constraint: func [spec] [
	ifn word? spec/1 [return 0xff]
	usable: add 0 reduce bind spec/2 class-constraint
	if eq? 'not spec/1 [
		usable: xor 0xff usable
	]
	usable
]

; Roles used in city blocks starting with NPC_TALKER_COMPANION
npc-roles: [
	companion	weaponsvendor	armorvendor
	foodvendor	tavernkeeper	reagentsvendor
	healer		innkeeper		guildvendor
	horsevendor	lordbritish		hawkwind
]

; Roles used in TMX files that match PersonNpcType
npc-roles2: [
	none
	talk		talk-beggar		talk-guard
	companion	weapons			armor
	food		tavern			reagents
	healer		inn				guild
	stable		lordbritish		hawkwind
]

weapon-flags: swap [
	0x01 lose
	0x02 losewhenranged
	0x04 choosedistance
	0x08 alwayshits
	0x10 magic
	0x40 attackthroughobjects
	0x80 absolute_range
   0x100 returns
   0x200 dontshowtravel
]

weapon-tiles: context [
	hittile: misstile: leavetile: none
	reset: does [hittile: misstile: leavetile: none]
]

dir-mask: [all 0x7e  W 2  N 4  E 8  S 16  adv 32  ret 64]
/*
direction-mask: [
	none	0x01
	west	0x02
	north	0x04
	east	0x08
	south	0x10
	advance	0x20
	retreat	0x40
	all		0x7e
]
*/

attribute-flags: func [blk spec] [
	flags: 0
	foreach [mask name] spec [
		if select blk name [flags: or flags mask]
	]
	flags
]

none-zero: func [v] [either v v 0]
alternate: func [v a] [either v v a]

;---------------------------------------
; TMX Map

tmx-fatal: func [file msg] [fatal data ["TMX" file '-' msg]]

; Convert <object> x & y pixel position to tile coord!.
tmx-position: func [x y] [
	x: to-int x
	y: to-int y
	div to-coord [x y] 16
]

object-parse: func [xml attr /local it] [
	blk: make block! 40
	; TMX 1.9 changed "type" to "class".
	parse construct xml ['=' ':'  "/>" '>'  "class=" "type:"] [some[
		thru "<object " tok: to '>' :tok (
			tok: to-block tok
			foreach it attr [
				append blk select tok it
			]
		)
	]]
	blk
]

; Convert TMX <object> to Boron #[role talk_id tile x y movement] values.
; The type attribute contains a movement char. & talk_id (e.g. "W 12")
tmx-npcs: func [object-tags /local gid x y name type] [
	ifn object-tags [return none]
	data: clear u16#[]
	attr: object-parse object-tags [gid x y name type]
	foreach [gid x y name type] attr [
		role: 0
		case [
			none? name []
			find name "role/" [
				role: enum-value npc-roles2 to-word skip name 5
			]
			find name "companion/" [role: 4]
		]
		talk_id: sub to-int skip type 2 1
		pos: tmx-position x y

		append data role
		appair data talk_id sub to-int gid 1
		appair data pos/1 sub pos/2 1
		append data enum-value "SWFA" type/1	; ObjectMovement
	]
	data
]

; Convert TMX <object name="" x="" y=""> to Boron (name x,y) values.
tmx-labels: func [object-tags /local name] [
	ifn object-tags [return none]
	blk: make block! 16
	attr: object-parse object-tags [name x y]
	foreach [name x y] attr [
		appair blk mark-sol to-word name tmx-position x y
	]
	blk
]

map-labels: none
load-tmx: func [file /extern width height data npcs map-labels /local it] [
	w: csv: none
	npcs-obj: labels-obj: none

	; See https://doc.mapeditor.org/en/stable/reference/tmx-map-format/
	parse read/text file [
		thru "<map "
		thru "<layer " thru {width="} w: thru {height="} h:
		thru "<data " thru {encoding="} enc: thru '>'
		csv: to "</data" :csv (
			w: to-int w
			h: to-int h
			enc: slice enc 3
		)
		any [
			thru "<objectgroup " thru {name="} group-name: to '"' :group-name
			objects: thru "</objectgroup>" :objects (
				switch group-name [
					"npcs"   [npcs-obj: objects]
					"labels" [labels-obj: objects]
				]
			)
		]
	]

	case [
		none? w			[tmx-fatal file "no map layer found"]
		none? csv		[tmx-fatal file "no map data found"]
		ne? enc "csv"	[tmx-fatal file "data encoding is not CSV"]
		not zero? and w 31 [tmx-fatal file "width is not a multiple of 32"]
		not zero? and h 31 [tmx-fatal file "height is not a multiple of 32"]
	]

	map-labels: tmx-labels labels-obj

	context [
		width: w
		height: h
		data: to-block construct csv [',' ' ']
		map it data [sub it 1]		; Convert indices to zero base.
		npcs: tmx-npcs npcs-obj
	]
]

; Return data as a series of chunks.
; Chunk-dim is the tile width & height of a single chunk.
map-chunks: func [tile-w tile-h data chunk-dim] [
	cdata: make binary! size? data
	chunks-y: div tile-h chunk-dim
	loop div tile-w chunk-dim [
		x: 0
		loop chunks-y [
			in-row: skip data mul ++ x chunk-dim
			loop chunk-dim [
				append cdata slice in-row chunk-dim
				in-row: skip in-row tile-w
			]
		]
		data: skip data mul chunk-dim tile-w
	]
	cdata
]

;---------------------------------------
; CDI Package

; The construct binary! rules for the Table Of Contents.
toc: []

chunk-count: 0
pkg-len: 0

cdi-begin: func [app_id /extern pkg-len] [
	clear toc
	write module-file bin: rejoin [#{DA7A7000} app_id "....----"]
	pkg-len: size? bin
]

cdi-end: does [
	toc: construct binary! toc
	write/append module-file toc

	; Poke toc offset & length.
	fh: open/write module-file
	write skip fh 8 construct binary! reduce [
		'u32 pkg-len size? toc
	]
	close fh
]

; Add toc entry and append data to pack-bin.
cdi-chunk: func [cdi-format name data /extern pkg-len] [
	append toc reduce [
		#{DA7A} 'big-endian 'u16 cdi-format
		to-binary name
		'little-endian 'u32
		pkg-len
		size? data
	]
	++ chunk-count
	write/append module-file data
	pkg-len: add pkg-len size? data
]

; Make CDI String Table (Form 1).
cdi-string-table1: func [dict] [
	store: make binary! 2048
	v16: make vector! 'u16
	foreach str dict [
		append v16 size? store
		ifn string? str [str: to-string str]
		appair store str '^0'
	]
	construct binary! reduce [
		'u8 1 0 'u16 'big-endian size? v16
		to-binary v16
		store
	]
]

audio-chunk: make block! 128

file-dict: make block! 128
file-id-seen: false

; File-id returns the zero-based index of str in file-dict.
; The global file-id-seen is set to true if str was already present in
; file-dict, or false if it was not.
file-id: func [str /extern file-id-seen] [
	if file-id-seen: find file-dict str [
		return index?/zero file-id-seen
	]
	idx: size? file-dict
	append file-dict str
	idx
]

pkg_id: "gyA^1"
img_id: "IM^0^0"
tmx_id: "MA^0^0"
npc_id: "NC^0^0"
mus_id: "MU^0^0"
snd_id: "SO^0^0"
sl_id:  "SL^0^0"
txf_id: "TF^0^0"

file_buf: make binary! 4096
module-layer: 0
stage-id: -1

poke-id: func [id n] [
	poke id 3 div n 256
	poke id 4 and n 255
]

; Return app_id of PNG or WebP chunk.
pack-img: func [path filename] [
	poke-id img_id add file-id filename module-layer

	ifn file-id-seen [
		fmt: select [
			".png"  0x1002
			".webp" 0x1006
		] file-ext filename
		cdi-chunk fmt img_id read/into join path filename file_buf
	]
	img_id
]

; Return app_id of map chunk.
pack-tmx: func [id filename chunk-dim] [
	poke-id tmx_id id

	tmx: load-tmx join root-path filename
	do bind [
		map-grid: either chunk-dim [
			map-chunks width height data chunk-dim
		][
			chunk-dim: 0
			append make binary! size? data data
		]

		either npcs [
			npc-count: div size? npcs 6
			npc-offset: size? map-grid
			npc-data: to-binary npcs
		][
			npc-count:
			npc-offset: 0
			npc-data: #{}
		]

		m1-data: construct binary! [
			u8  0x6D 0x01
			u16 width height
			u8  1 chunk-dim
			u16 0 npc-count
			u32 npc-offset
			map-grid
			npc-data
		]
	] tmx

	cdi-chunk 0x1FC0 tmx_id m1-data
	tmx_id
]

print-toc: does [
	print "table-of-contents: ["
	ascii: charset "0-9a-zA-Z"
	v32: #[]
	while [not tail? toc] [
		app_id: slice toc 4,4
		if parse app_id [4 ascii] [
			app_id: to-string app_id
		]

		append clear v32 slice toc 8,8
		print format ["  " 0 ' ' 11 -10 -9] [
			slice toc 4
			app_id
			to-hex v32/1
			v32/2
		]
		toc: skip toc 16
	]
	print ']'
]

;---------------------------------------

sl-out: sl-code: none
strip-shader: func [shader /extern sl-out sl-code] [
	ifn sl-out [
		sl-out: make binary! 4096
		sl-code: complement charset "^//"
	]
	clear sl-out
	emit-span: [(append sl-out slice span end) span:]
	parse span: shader [any[
		end: some sl-code
	  | "//" to '^/'   emit-span
	  | "/*" thru "*/" emit-span
	  | some '^/'      emit-span (append sl-out '^/')
	  | skip
	]]
	append sl-out slice span end
]

;---------------------------------------
; Load module configuration

config-file: join root-path %config.b
ifn exists? config-file [
	fatal noinput ["Cannot find config" config-file]
]

; Matches ModInfoValues in module.c.
modi: context [
	about:
	author:
	rules:
	version:
		none
]

; Matches ConfigValues in config_boron.c.
cfg: context [
	res-dict:
	res-spec:
	init-resources:
	stages:
	locales:
	ega-palette:
	creatures:
	terrain:
/*
	weapons:
	armors:
	weapons:
	tileanim:
	layouts:
	vendors:
*/
		none
]

module: func [blk block! /extern module-layer] [
	do bind blk modi
	mr: modi/rules
	if all [string? mr find mr '/'] [
		module-layer: 0x2000
	]
]

/*
  Parse data block! with rules and replace the data with a new block.
  The global 'blk variable is the output block which 'rules should append to.
*/
process-blk: func ['member data rules /extern blk] [
	blk: make block! 16
	ifn parse data [some rules] [
		fatal config join "Invalid " member
	]
	set in cfg member blk
]

include: func [file file!] [
	do join root-path file
]

resources: func [blk block!] [
	process-resources blk
]

init-resources: func [blk block!] [
	set in cfg 'init-resources process-resources blk
]

stages: func [sblk block! /extern stage-id] [
	high: add 1 last sort collect int! sblk
	stage-table: make block! high
	append/repeat stage-table none high
	foreach [stage-id it] sblk [
		if file? it [it: load join root-path it]
		parse it [some[
			pos:
			'init-resources block! (poke pos 2 process-resources pos/2)
		  | 'present skip block!   (poke pos 3 compile-present pos/2 pos/3)
		  | 'compile-gui block!    (change pos reduce [compile-gui pos/2 none])
		  | skip
		]]
		poke stage-table add 1 stage-id it
	]
	set in cfg 'stages stage-table
]

/*
map-proto: context [
    type: fname: size: chunk-dim:
    border: music:
    portals: labels: details:
        none
]
*/
map-portals: []

locales: func [spec block!] [
	process-blk locales sort/group spec 2 [
		tok: int! block! (
            at: /*make map-proto*/ tok/2

			clear map-portals
			map-roles: none

			fname: at/fname
			either eq? ".tmx" file-ext fname [
				pack-tmx tok/1 fname at/chunk-dim	; Sets map-labels
                at/labels: map-labels
				fname: none
			][
				fname: to-file fname
			]

			appair blk mark-sol fname to-coord reduce [
				tok/1
				enum-value [unset! world city shrine combat dungeon] at/type
				enum-value [wrap exit fixed] at/borderbehavior
				none-zero at/size/1
				none-zero at/size/2
				alternate at/size/3 1
			]
			append blk to-coord reduce [
				none-zero at/chunk-dim
				none-zero at/chunk-dim
				attribute-flags at [
					0x02 nolineofsight
					0x04 firstperson
				]
				none-zero at/music
			]
			append/block blk at/labels

			either at/portals [
				attrs: make block! 0
				parse at/portals [some [
					ptok: coord! word! coord! any [set-word! skip] :ptok (
						action: ptok/2
						if rad: ptok/retroActiveDest [
							rad: to-coord reduce [slice rad 1,3 rad/1]
						]

						append attrs mark-sol to-coord reduce [
							slice ptok/1 3		; coords
							slice ptok/3 1,3	; start
						]
						appair attrs ptok/condition to-coord reduce [
							none-zero ptok/3/1	; destid
							none-zero select [
								save-enter	1
								enter		1
								climb		2
								descend		4
								exit_north	8
								exit_east	0x10
								exit_south	0x20
								exit_west	0x40
							] action
							either eq? 'save-enter action 1 0	; saveLocation
							any [
								select [
									foot	1
									horse	2
									ship	4
									balloon	8
									footorhorse 3
								] ptok/transport
								1
							]
						]
						appair attrs rad ptok/message
					)
				  |	coord! 'moongate int! (
						append attrs mark-sol ptok/1
						appair attrs 'moongate ptok/3
						appair attrs none none
					)
				]]
				append/block blk sort/group attrs 5
			][
				append blk none
			]

			either at2: at/details [
				append/block blk switch at/type [
					city [
						tlk_name: at2/tlk_fname
						either eq? ".tlk" file-ext tlk_name [
							tlk_name: to-file tlk_name
						][
							tlk_name: copy pack-talk tlk_name
													 none-zero at2/voice
						]
						detail-blk: reduce [at2/name at2/type tlk_name]
						foreach [role id] at2/roles [
							append detail-blk to-coord reduce [
								enum-value npc-roles role
								id
							]
						]
						detail-blk
					]
					dungeon [
						reduce [at2/name at2/rooms]
					]
					shrine [
						reduce [at2/mantra enum-value [
							honesty compassion valor justice
							sacrifice honor spirituality humility
						] at2/virtue]
					]
					[fatal config ["Unexpected locale details for" at/type]]
				]
			][
				append blk none
			]
		)
	]
]

ega-palette: func [blk block!] [
	bin: make binary! mul 16 4
	foreach it blk [
		appair bin it/1 it/2
		appair bin it/3 255
	]
	set in cfg 'ega-palette bin
]

creatures: func [spec block! /extern blk] [
	process-blk creatures spec [
		tok: int! string! word! coord!
			any [word! opt block! | set-word! int!/lit-word!] :tok
		(append/block blk tok)
	]

	cstore: make block! 100
	blk: sort/field blk [1]
	cfg/creatures: map tok blk [
		begin: tail cstore
		append cstore mark-sol to-coord reduce [
			tok/1				; id
			none-zero tok/4/1	; hp
			none-zero tok/4/2	; xp
			none-zero tok/4/3	; encounterSize
			none-zero tok/leader
			none-zero tok/spawnsOnDeath
		]
		appair cstore tok/2 tok/3	; name & sprite

		at: skip tok 4
		append cstore to-coord reduce [
			attribute-flags at [
				  0x01 steals-food
				  0x02 steals-gold
				  0x04 casts-sleep
				  0x08 undead
				  0x10 good
				  0x20 swims		; water
				  0x20 sails		; water
				  0x40 cantattack	; nonAttackable
				  0x80 casts-negate
				0x0100 camouflage
				0x0200 wontattack	; noAttack
				0x0400 ambushes
			   ;0x0800 randomRanged
				0x1000 incorporeal
				0x2000 noChest
				0x4000 divides
				0x8000 spawnsOnDeath
			]
			attribute-flags at [
				  0x01 stationary
				  0x02 wanders
				  0x04 swims
				  0x08 sails
				  0x10 flies
				  0x20 teleports
				  0x40 canMoveOntoCreatures
				  0x80 canMoveOntoAvatar
				 0x100 forceOfNature
			]
			/*
			none-zero select [fire 1  sleep 2  poison 4] at/resists
			attribute-flags at [
				0x1 ranged
				0x2 leavestile
			]
			none-zero at/u4SaveId
			*/
		]

		/*
		ctiles: reduce [
			at/rangedHitTile
			at/rangedMissTile
			at/camouflageTile
			at/worldRangedTile
		]
		ifn empty? collect word! ctiles [
			append cstore ctiles
		]
		*/

		slice begin 4
	]
]

terrain: func [spec block! /extern blk] [
	process-blk terrain spec [
		tok: word! int!/coord! 11 word! skip (
			appair blk mark-sol tok/1 tok/2
			append blk to-coord reduce [
				0
				0
				enum-value [- sl vsl vvsl] tok/12 ; [fast slow vslow vvslow]
				0	; select [- 0 poison ] tok/13
				xor 0x7e none-zero select dir-mask tok/3
				xor 0x7e none-zero select dir-mask tok/4
			]
		)
	]
]

;---------------------------------------

context [
	istack: []
	icode: make block! 64
	op-accum:
	op-pos: 0

	begin-icode: does [
		append clear icode op-accum: 0
		op-pos: 1
	]

	end-icode: does [
		emit-op 255
		while [not zero? op-accum] [emit-op 0]
	]

	push-icode: func [blk /extern icode] [
		append/block clear istack icode
		appair istack op-accum op-pos
		icode: blk
		begin-icode
	]

	pop-icode: does [
		end-icode
		set [icode op-accum op-pos] istack
	]

	range?: func [val limits] [
		all [ge? val limits/1 le? val limits/2]
	]

	emit-op: func [op /extern op-accum op-pos] [
		op: mul op 0x1000000
		either zero? and 0xff op-accum [
			op-accum: or op div op-accum 256
		][
			poke icode op-pos mark-sol to-hex op-accum
			append icode 0
			op-pos: size? icode
			op-accum: op
		]
	]

	emit-op1: func [op int! arg] [emit-op op append icode arg]

	index0: func [list val] [
		index?/zero find list val
	]

	list-index0: func [list val] [
		if pos: find list val [
			return index?/zero pos
		]
		pos: size? list
		append list val
		pos
	]

	set 'compile-present func [dim prog] [
		vary-ctx: make block! 16
		res-ctx:  make block! 16
		render-store: make block! 4
		render-icode: make block! 64
		render-words: make block! 0
		append/block render-store render-words
		sprite-names: make block! 8     ; Maps names to index at compile time.
		sprite-counter: 0

		render-rule: [
			rtok: 'clear word! (
				emit-op1 1 select [
					color 0x4000 stencil 0x400 depth 0x100
					color+depth 0x4100 color+depth+stencil 0x4500
				] rtok/2
			)
		  | 'shader word! (
				emit-op1 2 list-index0 res-ctx rtok/2
			)
				any [rtok: option! word!/get-word! (
					either get-word? source: rtok/2 [
						emit-op1 3   list-index0 render-words rtok/1
						append icode list-index0 render-words source
					][
						emit-op1 select [
							screen-ocam 4 target-ocam 5 rclock 27 view-size 28
						] source
							list-index0 render-words rtok/1
					]
				)]
		  | 'texture int! word! (
				ifn range? tn: rtok/2 0,3 [
					fatal config ["Invalid texture number" tn]
				]
				emit-op1 add 6 tn list-index0 res-ctx rtok/3
			)
		  | 'varray word! (
				emit-op1 10 list-index0 res-ctx rtok/2
			)
		  | 'points int!/get-word! (
				either int? rtok: rtok/2 [
					emit-op1 11 rtok
				][
					emit-op1 12 index0 vary-ctx rtok
				]
			)
		  | 'blend word! (
				emit-op select [
					off 13 on 14 trans 15 add 16 burn 17
				] rtok/2
			)
		  | 'depth-test word! (
				emit-op select [off 18 on 19] rtok/2
			)
		  | 'draw-sprites   (emit-op 20)
		  | 'draw-tilemap   (emit-op 21)
		  | 'draw-text      (emit-op 22)
		  | 'draw-view-quad (emit-op 23)
		  | 'framebuf-view [int! (emit-op 24) | word! (
				emit-op1 25 list-index0 res-ctx rtok/2
			)]
		  | 'upscale-framebuf word! (
				emit-op1 26 list-index0 res-ctx rtok/2
			)
		]

		begin-icode

		ifn parse prog [some[
			tok: double! (
				emit-op1 1 mark-sol/clear to-int mul 1000.0 tok/1
			)
		  | 'clear-color double!/vec3! (
				emit-op1 2 to-vec3 tok/2
			)
		  | 'vary word! int!/double!/vec3! int!/double!/vec3! double! (
				emit-op select [int! 4 double! 5 vec3! 6] type? tok/3
				appair icode list-index0 vary-ctx to-set-word tok/2
					slice tok 2,3
			)
		  | 'play word! (
				emit-op1 7 list-index0 res-ctx tok/2
			)
		  | 'sprite word! coord! (
				emit-op1 8 list-index0 res-ctx tok/2
				pos: tok/3
				xpos: and pos/1 0x3fff
				ypos: and pos/2 0x3fff
				ifn zpos: pos/3 [zpos: 0]
				ifn range? zpos -7,8 [
					fatal config ["Invalid sprite depth" zpos]
				]
				append icode add mul 0x40000 xpos
							 add mul    0x10 ypos add zpos 7
				 ++ sprite-counter
			)
		  | 'image word! vec3! double! (
				emit-op1 9 list-index0 res-ctx tok/2
				pos: tok/3
				appair icode pos/1 pos/2
				append icode tok/4
			)
		  | 'render int!/word! into [
				(push-icode render-icode)
				some render-rule
				(pop-icode)
			](
				either word? slot: tok/2 [
					slot: 22	; render now
				][
					ifn range? slot 0,3 [
						fatal config ["Invalid render slot" slot]
					]
				]
				emit-op1 add 10 slot size? render-store
				append render-store mark-sol
					construct binary! slice render-icode -1
			)
		  | 'render-off int! (
				emit-op1 add 10 tok/2 0
			)
;		  | 'sprite-sheet word! (emit-op1 14 list-index0 res-ctx tok/2)
		  | 'tilemap word! word! coord! vec3! (
				emit-op1 14  list-index0 res-ctx tok/2
				appair icode list-index0 res-ctx tok/3 'u16
				appair icode tok/4 'u32
				pos: tok/5
				appair icode pos/1 pos/2
			)
		  | option! double! (
				emit-op1 select [
					slide-in-rt 15 slide-in-lt 16
					slide-in-up 17 slide-in-dn 18
					scale 37
				] tok/1 tok/2
			)
		  | option! vec3! (
				emit-op 19  ; /pq
				pos: tok/2
				appair icode pos/1 pos/2
			)
		  | option! int!/word! double! (
				opcode: select [frames 20 frames-loop 21] tok/1
				ifn opcode [fatal config ["Invalid sprite option" tok/1]]
				either int? count: tok/2 [
					frame-dur: to-int div mul 1000.0 tok/3 count
				][
					opcode: add 2 opcode
					count: list-index0 res-ctx count
					frame-dur: tok/3    ; Storing animation duration.
				]
				emit-op opcode
				appair icode count frame-dur
			)
		  | set-word! opt int!/double!/vec3! (
				value: tok/2
				opcode: select [int! 25 double! 26 vec3! 27] type? value
				either opcode [
					emit-op opcode
					appair icode list-index0 vary-ctx tok/1 value
				][
					if ne? 'sprite value [
						fatal config
							["Invalid variable assignment:" tok/1 value]
					]
					appair sprite-names tok/1 sprite-counter
				]
			)
		  | option! coord! coord! double! (
				switch tok/1 [
					morph-viewport [
						emit-op1 28 'u16
						appair icode slice tok 1,2 'u32
						append icode tok/4
					]
				]
			)
		  | option! (
				switch tok/1 [
					frame-jitter [emit-op 24]
				]
			)
		  | 'tilemap-effects word! word! (
				emit-op 29
				appair icode list-index0 res-ctx tok/2
							 list-index0 res-ctx tok/3
			)
		  | 'load-image word! word! (
				emit-op 30
				appair icode list-index0 res-ctx tok/2
							 list-index0 res-ctx tok/3
			)
		  | 'anykey (
				emit-op 31
			)
		  | 'font word! int!/double! (
				emit-op1 33 list-index0 res-ctx tok/2
				append icode to-double tok/3
			)
		  | 'text-box coord! word! opt [option! word!] (
				emit-op1 34 list-index0 res-ctx tok/3
				if ne? 3 size? tok/2 [
					fatal config "text-box coord requires X,Y,W"
				]
				append icode tok/2
				if eq? 'cursor tok/4 [
					emit-op1 35 select sprite-names tok/5
				]
			)
		  | 'set-text word! opt int! (
				emit-op1 36 list-index0 res-ctx tok/2
				append icode tok/3
			)
		]] [
			fatal config ["Invalid present program:" slice tok 3]
		]

		end-icode

		ifn empty? vary-ctx [
			vary-ctx: context append vary-ctx none
		]
		if eq? 1 size? render-store [render-store: none]

		;probe slice icode -1
		reduce [
			construct binary! slice icode -1
			vary-ctx
			res-ctx
			render-store
		]
	]
]

;---------------------------------------
; Compile GUI bytecode

context [
bcode: make binary! 256
data: make block! 10
gui-func: [
	emit: func [op] [append bcode op]
	emit2: func [op dat] [appair bcode op dat]

	layout-h:       func [blk block!] [emit 2 do blk emit 5]
	layout-v:       func [blk block!] [emit 3 do blk emit 5]
	layout-grid:    func [col int! blk block!] [emit2 4 col do blk emit 5]
	margin-p:       func [percent coord!] [emit 6 emit2 percent/1 percent/2]
	margin-ph:      func [percent int!] [emit2 7 percent]
	margin-pv:      func [percent int!] [emit2 8 percent]
	spacing-p:      func [percent int!] [emit2 9 percent]
	spacing-em:     func [em int!]      [emit2 10 em]
	fix-width-p:    func [percent int!] [emit2 11 percent]
	fix-width-em:   func [em int!]    [emit2 13 em]
	from-bottom:    does [emit 15]
	from-right:     does [emit 16]
	align-left:     does [emit 17]
	align-right:    does [emit 18]
	align-top:      does [emit 19]
	align-bottom:   does [emit 20]
	align-h-center: does [emit 21]
	align-v-center: does [emit 22]
	align-center:   does [emit 23]
	gap-p:          func [percent int!] [emit2 24 percent]
	grid-empty:     does [emit 25]
	font:           func [index int!] [emit2 26 index]
	font-size:      func [size int!]  [emit2 27 size]
	font-vsize:     func [size int!]  [emit2 28 size]
	font-color:     func [ci int!]    [emit2 29 ci]
	bg-color:       func [ci int! /semi-trans] [
						emit2 30 either semi-trans [add ci 128] ci
					]
	array-area:     func [wid int!]     [emit2 31 wid]
	button:         func [text string!] [emit 33 append data text]
	checkbox:       func [bit int!]     [emit2 34 bit]
	label:          func [text string!] [emit 35 append data text]
	list-dim:       func [dim coord!]   [emit2 36 first dim emit second dim]
	slider:         does [emit 37]
	min-width-em:   func [em int!]      [emit2 38 em]
	store-area:     does [emit 40]
]

set 'compile-gui func [spec block! /extern gui-func] [
	if block? gui-func [
		gui-func: context gui-func
	]
	clear bcode
	clear data
	do bind spec gui-func
	append bcode 0
	ifn zero? and 1 size? bcode [
		append bcode 0  ; Pad to align idx.
	]

	;probe data
	str: make binary! 128
	idx: make vector! 'u16
	append idx size? data
	foreach it data [
		append idx size? str
		appair str it 0
	]

	off-idx: add 8 size? bcode
	off-str: add off-idx mul 2 size? idx
	construct binary! reduce [
		to-binary "GUI^0"
		'u16 off-idx off-str
		bcode
		to-binary idx
		str
	]
]
]

;---------------------------------------
; Resource Banks

resource-banks: context resource-types: [
	; Order matches ResourceType in config.h
	music: sound: font: image: texture: framebuf: varray: shader: tilemap:
	0,0,0,0
]
resource-dict: []
resource-def: []
resource-list: []
resource-init: []
rdefA: rdefB: none

push-rdef: func [blk /extern rdefA rdefB] [
	rdefA: size? resource-def
	append resource-def blk
	rdefB: size? resource-def
]

process-resources: func [blk /local dt bank desc /extern rdefA rdefB] [
	ifn parse blk [some [word! int! block!]] [
		fatal config "Invalid resources"
	]

	res-names: make block! 32
	foreach [dt bank desc] blk [
		collect/into set-word! desc res-names

		switch dt [
			music [
				app_id: copy mus_id
				path: root-path
				parse desc [some [
					tok: set-word! file! (
						fn: join path tok/2
						either pos: find audio-chunk fn [
							ci: pos/2
						][
							ci: chunk-count
							poke-id app_id add div size? audio-chunk 2
												module-layer
							emit-audio-chunk app_id fn
						]
						appair resource-dict tok/1
								to-coord [0 bank stage-id ci -1]
					)
				  | 'path file! (path: terminate join root-path tok/2 '/')
				; | 'parts file! ()
				]]
			]
			sound [
				app_id: copy snd_id
				path: root-path
				parse desc [some [
					tok: set-word! file! (
						fn: join path tok/2
						either pos: find audio-chunk fn [
							ci: pos/2
						][
							ci: chunk-count
							poke-id app_id add div size? audio-chunk 2
												module-layer
							emit-audio-chunk app_id fn
						]
						appair resource-dict tok/1
								to-coord [1 bank stage-id ci -1]
					)
				  | 'path file! (path: terminate join root-path tok/2 '/')
				]]
			]
			font [
				app_id: copy txf_id
				path: join root-path %font/
				parse desc [some [
					tok: set-word! file! (
						fn: tok/2
						poke-id app_id add file-id fn module-layer
						ifn file-id-seen [
							ci: chunk-count
							cdi-chunk 0x5FC0 app_id
								read/into join path fn file_buf
							appair resource-dict tok/1
									to-coord [2 bank stage-id ci -1]
						]
					)
				]] [fatal config ["Invalid font definition:" tok/1]]
			]
			image [
				app_id: copy img_id
				ifn parse desc [some [
					tok: set-word! ['u4lzw | 'u4raw] coord! file!
							opt block! :tok (
						push-rdef next tok
						appair resource-dict tok/1
								to-coord [3 bank stage-id -1 rdefA rdefB]
					)
				  | set-word! file! opt block! :tok (
						ci: chunk-count
						pack-img root-path tok/2
						either block? tok/3 [
							push-rdef skip tok 2
						][
							rdefA: rdefB: -1
						]
						appair resource-dict tok/1
								to-coord [3 bank stage-id ci rdefA rdefB]
					)
				]] [fatal config ["Invalid image definition:" tok/1]]
			]
			texture [
				ifn parse desc [some [
					tok:
					set-word! ['atlas coord! block! | opt 'depth coord!]
							any option!/int! :tok (
						push-rdef skip tok 1
						appair resource-dict tok/1
								to-coord [4 bank stage-id -1 rdefA rdefB]
					)
				   | set-word! file! any option!/int!/block! :tok (
						ci: chunk-count
						pack-img root-path tok/2
						either eq? option!/block! type? tok/3 [
							push-rdef skip tok 2
						][
							rdefA: rdefB: -1
						]
						appair resource-dict tok/1
								to-coord [4 bank stage-id ci rdefA rdefB]
					)
				]] [fatal config ["Invalid texture definition:" tok/1]]
			]
			framebuf [
				parse desc [some [
					tok: set-word! some word! :tok (
						push-rdef skip tok 1
						appair resource-dict tok/1
								to-coord [5 bank stage-id -1 rdefA rdefB]
					)
				]]
			]
			varray [
				parse desc [some [
					tok: set-word! int!/coord! [
						'tris  int!
							(push-rdef reduce [tok/2 mul 3 tok/4])
					  | 'quads int!
							(push-rdef reduce [tok/2 mul -4 tok/4])
					  | int!/word! (push-rdef slice tok 1,2)
					] (
						appair resource-dict tok/1
								to-coord [6 bank stage-id -1 rdefA rdefB]
					)
				]]
			]
			shader [
				app_id: copy sl_id
				path: join root-path %shader/
				option-end: none
				ifn parse desc [some [
					tok: set-word! file! any [
						option! [int!/vec3! | word! opt coord!] option-end:
					] (
						fn: tok/2
						poke-id app_id add file-id fn module-layer
						ifn file-id-seen [
							ci: chunk-count
							cdi-chunk 0x0001 app_id
								strip-shader read/into join path fn file_buf
							either option-end [
								push-rdef slice skip tok 2 option-end
								option-end: none
							][
								rdefA: rdefB: -1
							]
							appair resource-dict tok/1
									to-coord [7 bank stage-id ci rdefA rdefB]
						]
					)
				]] [fatal config ["Invalid shader definition:" tok/1]]
			]
			tilemap [
				ifn parse desc [some [
					tok: set-word! coord! word!/file! (
						push-rdef slice tok 1,2
						appair resource-dict tok/1
								to-coord [8 bank stage-id -1 rdefA rdefB]
					)
				  | set-word! 'none (
						appair resource-dict tok/1
								to-coord [8 bank stage-id -1 -1]
					)
				]] [fatal config ["Invalid tilemap definition:" tok/1]]
			]
		]
	]
	res-names
]

;---------------------------------------
; NPC Talk

context [
store: sdict: none
meld: func [data] [
	map it data [
		switch type? it [
			string! [
				ifn ss: pick sdict it [
					ss: tail store
					append store it
					poke sdict it ss: slice ss tail store
					append store '^0'
				]
				it: ss
			]
			block! [meld it]
		]
		it
	]
]

set 'meld-strings func [data /extern store sdict] [
	store: make string! 1024
	sdict: make hash-map! 256
	meld data
	sdict
]
]

npc-talk: context [
	name:
	pronoun:
	look: none
	turn-away: 0
	topics: none
]

; Return number of spoken lines for NPC.
spoken-lines: func [topics] [
	count: 4    ; Matches VP_KEYWORD
	parse topics [some[
		string! string! (++ count)
	  | word! string! block! (count: add count 3)
	]]
	count
]

; Return app_id of talk chunk.
pack-talk: func [filename voice-stream /local it] [
	poke-id npc_id file-id filename
	ifn file-id-seen [
		meld-strings spec: load join root-path filename
		tblk: make block! mul size? spec 5
		voice-line: 0
		foreach it spec [
			npc-blk: values-of make npc-talk it
			turn-away: pick npc-blk 4
			poke npc-blk 4 to-coord [turn-away voice-stream voice-line]
			append tblk mark-sol npc-blk
			voice-line: add voice-line spoken-lines last npc-blk
		]
		if ge? verbose 2 [
			print [filename '>' to-binary npc_id]
			probe tblk
		]
		cdi-chunk 0x4007 npc_id serialize tblk
	]
	npc_id
]

;---------------------------------------
; Build module package.

emit-audio-chunk: func [app_id fname] [
	fmt: select [
		".wav"  0x2006
		".mp3"  0x2007
		".ogg"  0x2008
		".flac" 0x2011
		".rfx"  0x2030
	] ext: file-ext fname
	ifn fmt [fatal config ["Unknown audio file extension" ext]]

	appair audio-chunk fname chunk-count
	cdi-chunk fmt app_id read/into fname file_buf
]

/*
process-sound: func [blk app_id] [
	parts: make block! 16
	path: root-path
	high: n: 0
	app_id: copy app_id
	parse blk [some[
		tok: set-word! file! (
			filename: tok/2
			poke-id app_id add file-id filename module-layer
			either file-id-seen [
				append bchunk-index -1	; TODO: Find existing chunk number.
			][
				append bchunk-index chunk-count
				emit-audio-chunk app_id join path filename
			]
		)
	  | 'path file! (path: terminate join root-path tok/2 '/')
	  | 'parts file! (
			fn: join path tok/2

			append bchunk-index chunk-count
			poke-id app_id ++ n
			emit-audio-chunk app_id fn

			vec: make vector! 'f32
			append vec collect double!
						load construct to-string fn [".ogg" "-parts.b"]
			appair parts n vec
			high: maximum high n
		)
	]]

	either empty? parts n [
		append/repeat parts-table: make block! high none high
		foreach [n vec] parts [
			poke parts-table n vec
		]
		parts-table
	]
]
*/

/*
  Parse data block! with rules but keep the data.
*/
/*
process-minimal: func ['data rules] [
	ifn orig: get data [return none]

	ifn parse orig [some rules] [
		fatal config join "Invalid " data
	]
	orig
]
*/

transforms: make binary! 512
current-trans: none
new-transform: func [type data /extern current-trans] [
	current-trans: tail transforms
	either coord? data [
		x: data/1
		y: data/2
		w: data/3
		h: data/4
	][
		x: data
		y: w: h: 0
	]
	append transforms construct binary! [	; struct TileAnimTransform
		u8  type 0 0 0
		u16 x y w h
		u32 0 0
	]
]

cdi-begin pkg_id

do config-file
ifn modi/rules [
	fatal config "Missing module/rules"
]

;res-banks: resource-banks
cfg/res-dict: context resource-dict
cfg/res-spec: resource-def

/*
process-cfg [
	process-blk armors [
		tok: string! int! opt [word! block!] (
			appair blk tok/1 to-coord reduce [
				tok/2
				make-constraint skip tok 2
			]
		)
	]

	process-blk weapons [
		   ; Abbr.   Name   Range Damage
		tok: string! string! int! int!
		(constr: 0xff flags: 0 weapon-tiles/reset) any [
			stok:
			word! block!	(constr: make-constraint stok)
		  | set-word! word!	(set in weapon-tiles stok/1 stok/2)
		  | word!			(flags: or flags select weapon-flags stok/1)
		] (
			append blk slice tok 2
									  ;[range  damage  flags  canuse]
			append blk to-coord reduce [tok/3  tok/4  flags  constr]
			val: slice values-of weapon-tiles 3
			ifn empty? collect word! val [
				append blk val
			]
		)
	]

	layout-subimages: func [spec block! out block! width none!/int!] [
		; Stack vertically if width not provided.
		if none? width [width: 0]

		pos: 0,0
		size: 16,16
		next-tile: [
			x: add pos/1 size/1
			y: pos/2
			if ge? x width [
				x: 0
				y: add y size/2
			]
			pos: to-coord [x y]
		]
		emit-name: [append out mark-sol tok/1]

		parse spec [some[
			tok:
			'at   set pos  coord!
		  | 'size set size coord!
		  | word! coord! (do emit-name append out tok/2)
		  | 'image-width set width int!
		  | word! int! (
			  frames: tok/2
			  do emit-name append out to-coord [pos size frames]
			  do next-tile
			  loop sub frames 1 [
				  appair out mark-sol '_cel to-coord [pos size]
				  do next-tile
			  ]
			)
		  | word! (
			  do emit-name append out to-coord [pos size]
			  do next-tile
			)
		]]
	]

	process-img: func [name at subimages] [
		fname: at/filename
		either find [".png" ".webp"] file-ext fname [
			fname: copy pack-img image-path fname
			ftype: 'png
		][
			ftype: at/filetype
		]

		appair blk mark-sol name fname
		appair blk to-coord reduce [
			alternate at/width  -1
			alternate at/height -1
			alternate at/depth  -1
		] to-coord reduce [
			enum-value [
				none png u4raw u4rle u4lzw
				u5lzw fmtowns fmtowns-pic fmtowns-tif
			] ftype
			; Skip name which can be 'tiles (assuming name is first).
			none-zero select skip at 2 'tiles
			enum-value [
				none intro abyss abacus dungns transparent0
				blackTransparencyHack fmtownsscreen
			] at/fixup
		]

		either block? subimages [
			append/block blk make block! 8
			layout-subimages subimages last blk at/width
		][
			append blk none
		]
	]

	image-path: join root-path %image/
	n: 0

	process-blk graphics [
		'image set at paren! tok: opt block! (
			process-img at/name at tok/1
		)
	  | tok: set-word! 'atlas coord! block! (
			append blk reduce [
				mark-sol to-word tok/1 'atlas tok/3 tok/4 none
			]
		)
	  | set-word! string! opt block! (
			poke tmp-attr: [filename: none] 2 tok/2
			process-img to-word tok/1 tmp-attr tok/3
		)
	]

	; Use draw-lists as is except remove any string! from quads block.
	process-minimal draw-lists [
		word! coord! tok: block! (
			poke tok 1 collect coord! tok/1
		)
	]

	process-blk tileanim [
		tok: set-word! (
			aspec: tok/2
			anim-chance: 0
			if eq? 'random aspec/1 [anim-chance: aspec/2]

			appair blk
				to-word tok/1
				to-coord reduce [
					div size? transforms 20	 ; sizeof(TileAnimTransform)
					anim-chance
				]
			current-trans: none
		) into [some [
			'random set n int! (
				; Ignore initial 'random as it is handled above.
				if current-trans [poke current-trans 2 n]
			)
		  | 'invert set n coord!	  (new-transform 0 n)
		  | 'scroll set n int!/coord! (new-transform 1 n)
		  | 'frame					  (new-transform 2 0)
		  | 'pixel_color set n coord! (new-transform 3 n) into [
				set colA coord! set colB coord! (
					n: 13
					foreach chan reduce [
						colA/1 colA/2 colA/3 0
						colB/1 colB/2 colB/3 0
					][
						poke current-trans ++ n chan
					]
				)
			]
		  | 'context-frame set n int! (
				poke current-trans 3 1  ; ACON_FRAME
				poke current-trans 4 n
			)
		]]
	]
	ifn empty? transforms [
		append blk mark-sol transforms
	]

	process-blk tile-rules [
		set at paren! (
			appair blk mark-sol at/name to-coord reduce [
				enum-value [fast slow vslow vvslow]
					at/speed
				enum-value [none fire sleep poison poisonfield electricity lava]
					at/effect
				xor 0x7e none-zero select direction-mask at/cantwalkon
				xor 0x7e none-zero select direction-mask at/cantwalkoff
				attribute-flags at [
					  0x01 ship
					  0x02 horse
					  0x04 balloon
					  0x08 dispel
					  0x10 talkover
					  0x20 door
					  0x40 lockeddoor
					  0x80 chest
					0x0100 canattackover
					0x0200 canlandballoon
					0x0400 replacement
					0x0800 onWaterOnlyReplacement
					0x1000 foreground
					0x2000 livingthing
				]
				attribute-flags at [		; movement
					1 swimable
					2 sailable
					4 unflyable
					8 creatureunwalkable
				]
			]
		)
	]

	process-blk tileset [
		set at paren! (
			appair blk mark-sol at/name at/rule
			appair blk at/image at/animation
			appair blk at/directions to-coord reduce [
				none-zero at/frames
				none-zero select [square 1 round 2] at/opaque
				attribute-flags at [
					2 usesReplacementTileAsBackground
					4 usesWaterReplacementTileAsBackground
					8 tiledInDungeon
				]
			]
		)
	]
]
*/

if ge? verbose 2 [
	probe modi
	probe cfg
	probe file-dict
]

cdi-chunk 0x0006 "MODI" cdi-string-table1 values-of modi
cdi-chunk 0x7FC0 "CONF" serialize reduce [cfg]
ifn empty? file-dict [
	cdi-chunk 0x0006 "FNAM" cdi-string-table1 file-dict
]
cdi-end

if ge? verbose 1 [print-toc]
